# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from django.conf.urls import url

from .universe.views import UniverseView

urlpatterns = [
    url(r'pulp_cookbook/market/(?P<path>.+)/universe$', UniverseView.as_view())
]



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from gettext import gettext as _

from django.db import transaction
from django_filters.rest_framework import filterset
from rest_framework.decorators import detail_route
from rest_framework import serializers, status
from rest_framework.response import Response

from pulpcore.plugin.models import Artifact, Repository, RepositoryVersion
from pulpcore.plugin.tasking import enqueue_with_reservation

from pulpcore.plugin.viewsets import (
    ContentViewSet,
    OperationPostponedResponse,
    PublisherViewSet)
from rest_framework_nested.relations import NestedHyperlinkedRelatedField

from . import tasks
from .models import CookbookPackageContent, CookbookPublisher
from .serializers import (
    CookbookPackageContentSerializer,
    CookbookPublisherSerializer)

from pulp_cookbook.metadata import CookbookMetadata


class CookbookPackageContentFilter(filterset.FilterSet):
    class Meta:
        model = CookbookPackageContent
        fields = [
            'name',
            'version'
        ]


class _RepositoryPublishURLSerializer(serializers.Serializer):

    repository = serializers.HyperlinkedRelatedField(
        help_text=_('A URI of the repository to be published.'),
        required=False,
        label=_('Repository'),
        queryset=Repository.objects.all(),
        view_name='repositories-detail',
    )

    repository_version = NestedHyperlinkedRelatedField(
        help_text=_('A URI of the repository version to be published.'),
        required=False,
        label=_('Repository Version'),
        queryset=RepositoryVersion.objects.all(),
        view_name='versions-detail',
        lookup_field='number',
        parent_lookup_kwargs={'repository_pk': 'repository__pk'},
    )

    def validate(self, data):
        repository = data.get('repository')
        repository_version = data.get('repository_version')

        if not repository and not repository_version:
            raise serializers.ValidationError(
                _("Either the 'repository' or 'repository_version' need to be specified"))
        elif not repository and repository_version:
            return data
        elif repository and not repository_version:
            version = RepositoryVersion.latest(repository)
            if version:
                new_data = {'repository_version': version}
                new_data.update(data)
                return new_data
            else:
                raise serializers.ValidationError(
                    detail=_('Repository has no version available to publish'))
        raise serializers.ValidationError(
            _("Either the 'repository' or 'repository_version' need to be specified "
              "but not both.")
        )


class CookbookPackageContentViewSet(ContentViewSet):
    endpoint_name = 'cookbook/cookbooks'
    queryset = CookbookPackageContent.objects.all()
    serializer_class = CookbookPackageContentSerializer
    filter_class = CookbookPackageContentFilter

    @transaction.atomic
    def create(self, request):
        data = request.data
        try:
            artifact = self.get_resource(data['artifact'], Artifact)
        except KeyError:
            raise serializers.ValidationError(detail={'artifact': _('This field is required')})

        try:
            metadata = CookbookMetadata.from_cookbook_file(artifact.file.name, data['name'])
        except KeyError:
            raise serializers.ValidationError(detail={'name': _('This field is required')})
        except FileNotFoundError:
            raise serializers.ValidationError(
                detail={'artifact': _('No metadata.json found in cookbook tar')})

        try:
            if data['version'] != metadata.version:
                raise serializers.ValidationError(
                    detail={'version': _('version does not correspond to version in cookbook tar')})
        except KeyError:
            pass
        data['version'] = metadata.version
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        content = serializer.save(dependencies=metadata.dependencies)
        content.artifact = artifact

        headers = self.get_success_headers(request.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class CookbookPublisherViewSet(PublisherViewSet):
    endpoint_name = 'cookbook'
    queryset = CookbookPublisher.objects.all()
    serializer_class = CookbookPublisherSerializer

    @detail_route(methods=('post',), serializer_class=_RepositoryPublishURLSerializer)
    def publish(self, request, pk):
        """
        Publishes a repository. Either the ``repository`` or the ``repository_version`` fields can
        be provided but not both at the same time.
        """
        publisher = self.get_object()
        serializer = _RepositoryPublishURLSerializer(data=request.data,
                                                     context={'request': request})
        serializer.is_valid(raise_exception=True)
        repository_version = serializer.validated_data.get('repository_version')

        result = enqueue_with_reservation(
            tasks.publish,
            [repository_version.repository, publisher],
            kwargs={
                'publisher_pk': str(publisher.pk),
                'repository_version_pk': str(repository_version.pk)
            }
        )
        return OperationPostponedResponse(result, request)



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import tarfile
import json


class CookbookMetadata:
    """
    Represents metadata extracted from a cookbook tar archive.

    Attributes:
        metadata (dict): content of the 'metadata.json' file of a cookbook
    """

    def __init__(self, metadata):
        self.metadata = metadata

    @property
    def name(self):
        return self.metadata['name']

    @property
    def version(self):
        return self.metadata['version']

    @property
    def dependencies(self):
        return self.metadata['dependencies']

    @classmethod
    def from_cookbook_file(cls, file_name, name):
        """
        Construct a CookbookMetadata instance from a cookbook tar archive.

        Args:
            file_name (str): filename of the cookbook tar archive
            name (str): name of the cookbook ("metadata.json" file
                        is expected to be in the directoy `<name>`)

        Returns:
            CookbookMetadata: Instance containing the extracted metadata
        """
        tf = tarfile.open(file_name)
        for element in tf:
            if element.isfile() and element.name == name + '/metadata.json':
                metadata = json.load(tf.extractfile(element))
                # TODO: check name consistency, raise error
                return CookbookMetadata(metadata)
        raise FileNotFoundError


class Entry:
    """
    Universe entry: info about a cookbook in the universe file.


    Attributes:
        name (str): cookbook name
        version (str): cookbook version
        download_url (str): URL of cookbook tar package
        dependencies (dict): cookbook dependencies
    """

    def __init__(self, name, version, download_url, dependencies):
        """
        Args:
        """
        self.name = name
        self.version = version
        self.download_url = download_url
        self.dependencies = dependencies

    @property
    def data(self):
        return {
            'location_type': 'uri',
            'location_path': self.download_url,
            'download_url': self.download_url,
            'dependencies': self.dependencies
        }


class Universe:
    """
    Represents the cookbook universe.
    Describes cookbooks contained within the directory.

    Attributes:
        relative_path (str): An relative path to the universe.
    """

    def __init__(self, relative_path):
        """
        Args:
            relative_path (str): An relative path to the universe.
        """
        self.relative_path = relative_path

    def write(self, entries):
        """
        Write the universe JSON file.

        Args:
            entries (iterable): The entries to be written.
        """
        universe = dict()
        for entry in entries:
            try:
                versions = universe[entry.name]
            except KeyError:
                universe[entry.name] = versions = dict()
            versions[entry.version] = entry.data
        with open(self.relative_path, 'w+') as fp:
            json.dump(universe, fp)



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from rest_framework import serializers

from pulpcore.plugin.models import Artifact
from pulpcore.plugin.serializers import ContentSerializer, RemoteSerializer, PublisherSerializer

from .models import CookbookPackageContent, CookbookRemote, CookbookPublisher


class CookbookPackageContentSerializer(ContentSerializer):
    name = serializers.CharField(
        help_text="name of the cookbook"
    )
    version = serializers.CharField(
        help_text="version of the cookbook",
        required=False
    )
    dependencies = serializers.JSONField(
        help_text="dependencies of the cookbook",
        read_only=True
    )
    artifact = serializers.HyperlinkedRelatedField(
        view_name='artifacts-detail',
        help_text="tar archive containing the cookbook",
        queryset=Artifact.objects.all()
    )
    artifacts = None

    class Meta:
        fields = tuple(field for field in ContentSerializer.Meta.fields if field != 'artifacts') \
            + ('name', 'version', 'dependencies', 'artifact')
        model = CookbookPackageContent


class CookbookRemoteSerializer(RemoteSerializer):

    class Meta:
        fields = RemoteSerializer.Meta.fields
        model = CookbookRemote


class CookbookPublisherSerializer(PublisherSerializer):
    class Meta:
        fields = PublisherSerializer.Meta.fields
        model = CookbookPublisher



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
import os

from gettext import gettext as _

from django.core.files import File

from pulpcore.plugin.models import (
    RepositoryVersion,
    Publication,
    PublishedArtifact,
    PublishedMetadata)
from pulpcore.plugin.tasking import WorkingDirectory

from pulp_cookbook.app.models import CookbookPackageContent, CookbookPublisher
from pulp_cookbook.metadata import Entry, Universe
from pulp_cookbook.app.universe.views import path_template

log = logging.getLogger(__name__)


def publish(publisher_pk, repository_version_pk):
    """
    Use provided publisher to create a Publication based on a RepositoryVersion.

    Args:
        publisher_pk (str): Use the publish settings provided by this publisher.
        repository_version_pk (str): Create a publication from this repository version.
    """
    publisher = CookbookPublisher.objects.get(pk=publisher_pk)
    repository_version = RepositoryVersion.objects.get(pk=repository_version_pk)

    log.info(
        _('Publishing: repository=%(repository)s, version=%(version)d, publisher=%(publisher)s'),
        {
            'repository': repository_version.repository.name,
            'version': repository_version.number,
            'publisher': publisher.name,
        })

    with WorkingDirectory():
        with Publication.create(repository_version, publisher) as publication:
            universe = Universe('__universe__')
            universe.write(populate(publication))
            metadata = PublishedMetadata(
                relative_path=os.path.basename(universe.relative_path),
                publication=publication,
                file=File(open(universe.relative_path, 'rb')))
            metadata.save()

    log.info(
        _('Publication: %(publication)s created'),
        {
            'publication': publication.pk
        })


def populate(publication):
    """
    Populate a publication.

    Create published artifacts and yield a Universe Entry for each.

    Args:
        publication (pulpcore.plugin.models.Publication): A Publication to populate.

    Yields:
        Entry: Universe entry for each cookbook
    """
    for content in CookbookPackageContent.objects.filter(
            pk__in=publication.repository_version.content).order_by('-created'):
        relative_path = 'cookbook_files/{}/{}/'.format(content.name,
                                                       content.version.replace('.', '_'))
        for content_artifact in content.contentartifact_set.all():
            art_path = os.path.join(relative_path, content_artifact.relative_path)
            published_artifact = PublishedArtifact(
                relative_path=art_path,
                publication=publication,
                content_artifact=content_artifact)
            published_artifact.save()
            entry = Entry(
                name=content.name,
                version=content.version,
                download_url=path_template(art_path),
                dependencies=content.dependencies)
            yield entry



#!/usr/bin/env python3

from setuptools import setup, find_packages

requirements = [
    'pulpcore-plugin>=0.1.0b2',
]

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='pulp-cookbook',
    version='0.0.1a1',
    description='Cookbook plugin for the Pulp Project',
    long_description=long_description,
    author='Simon Baatz',
    author_email='gmbnomis@gmail.com',
    url='https://github.com/gmbnomis/pulp_cookbook/',
    install_requires=requirements,
    include_package_data=True,
    packages=find_packages(exclude=['test']),
    classifiers=(
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ),
    entry_points={
        'pulpcore.plugin': [
            'pulp_cookbook = pulp_cookbook:default_app_config',
        ]
    }
)



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from django.db import models

from pulpcore.plugin.models import Content, ContentArtifact, Remote, Publisher
from pulpcore.plugin.fields import JSONField


class CookbookPackageContent(Content):
    """
    The "cookbook" content type.

    Content of this type represents a Chef cookbook uniquely
    identified by name and version.

    Fields:
        name (str): The name of the cookbook.
        version (str): The version of the cookbook.
        dependencies (JSON): The dependency information of the cookbook.
    """

    TYPE = 'cookbook'

    name = models.TextField(blank=False, null=False)
    version = models.TextField(blank=False, null=False)
    dependencies = JSONField(blank=False, null=False)

    @property
    def artifact(self):
        return self.artifacts.get().pk

    @artifact.setter
    def artifact(self, artifact):
        if self.pk:
            ca = ContentArtifact(artifact=artifact,
                                 content=self,
                                 relative_path="{}-{}.tar.gz".format(self.name, self.version))
            ca.save()

    class Meta:
        unique_together = (
            'name',
            'version'
        )


class CookbookRemote(Remote):
    """
    Remote for "cookbook" content.
    """
    TYPE = 'cookbook'


class CookbookPublisher(Publisher):
    """
    Publisher for "cookbook" content.
    """
    TYPE = 'cookbook'



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from pulpcore.plugin import PulpPluginAppConfig


class PulpCookbookPluginAppConfig(PulpPluginAppConfig):
    name = 'pulp_cookbook.app'
    label = 'pulp_cookbook'
